(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_io.mli 239 2010-06-23 16:49:03Z gerd $ *)

(** Utility library for record-based I/O *)

(** A file consists here of lines. Each line is terminated by LF, and
    is considered as a record.

    The record reader can be used to iterate over a whole file, or
    only a part. For the latter, it is assumed that the file is processed
    block by block. Of course, it is then possible that the lines do not
    end at block boundaries. However, it must be determined whether the
    block reader for block N or the block reader for block N+1 processes
    such lines. The following rules do that:

    - The first block does not have this problem. Its first line is always
      processed by the block reader for block 0.
    - For the other blocks we define that the first processed line
      is the line starting after the first LF character in the block.
    - For all blocks we define that the last processed line is the
      line following the last LF character in the block. The line can
      be stored partly, even fully in the next block. The block reader
      has to read the next block to allow this.
    - A block must contain at least one LF character (and hence, have
      at least one line to process)

    For best efficiency, the block reader should not be used for
    reading individual blocks, but for contiguous ranges of blocks.
 *)

class type record_reader =
object
  method pos_in : int
    (** The ordinal number of the record that will be read next. Numbers
	start at 0
     *)

  method input_record : unit -> string
    (** Reads one record, and advances the cursor. May raise [End_of_file] *)

  method peek_record : unit -> string
    (** Reads the record, but does not advance the cursor, i.e. the
	next [peek_record] or [input_record] will read the line again.
	May raise [End_of_file].
     *)

  method close_in : unit -> unit
    (** Releases resources (e.g. closes transactions) *)

  method abort : unit -> unit
    (** Drops resources - intended to be used for error cleanup *)
end


class type record_writer =
object
  method pos_out : int
    (** The ordinal number of the record that will be written next. Numbers
	start at 0
     *)

  method bpos_out : int64
    (** The byte position of the record that will be written next *)

  method output_record : string -> unit
    (** Outputs these records. The records must not contain LF chars.
	The length is limited to the blocksize-1.
     *)

  method flush : unit -> unit
    (** Flushes the records from the buffer to disk *)

  method close_out : unit -> unit
    (** Releases resources (e.g. closes transactions) *)

  method abort : unit -> unit
    (** Drops resources - intended to be used for error cleanup *)
end


val read_file : Plasma_client.plasma_cluster -> 
                string -> int64 -> int64 -> record_reader
  (** [read_file c name block len]: Reads from [name], starting at [block],
      ending at [block+len-1]. Reading is done in a separate transaction.

      Note that len>=1 is a requirement here.

      The function configures the number of buffers of [c].

      The cluster [c] is set to aborted state when not used. Note that this
      also affects all transactions unrelated to [read_file], so is best
      to create a separate [plasma_cluster] object for reading.
   *)

val read_multiple : (unit -> record_reader) list -> record_reader
  (** Constructs a record reader that reads from the input readers one
      after the other
   *)

val write_file : Plasma_client.plasma_cluster -> string -> record_writer
  (** [write_file name]: Appends records to this file (which must already
      exist). Writing is done in separate transactions.

      The function configures the number of buffers of [c].

      As [read_file], the cluster [c] is set to aborted state when not used.
      Note that this also affects all transactions unrelated to [read_file],
      so is best to create a separate [plasma_cluster] object for writing.
   *)

val write_multiple : Plasma_client.plasma_cluster -> string -> int64 ->
                     (string -> int -> string) ->
                        record_writer
  (** [write_multiple c prefix limit create]: Writes into a sequence of files
      whose names are composed of [prefix] followed by an integer [k]. The
      files are created by calling [create prefix k]. A new file is started
      when the current file reaches the size [limit] (in bytes).
   *)
			 

val create_file : ?repl:int -> Plasma_client.plasma_cluster -> string -> unit
  (** [create_file c name]: Creates this file exclusively. [repl] is
      the replication factor, 0 by default (i.e. use server default).
   *)


val delete_file :  Plasma_client.plasma_cluster -> string -> unit
  (** Delete this file *)

val file_blocks : Plasma_client.plasma_cluster -> string -> int64
  (** Get the length of the file in blocks *)
